__all__ = ['ClientException', 'ValidationException']


class ClientException(Exception):
    """Raised when the server does something we don't expect."""

    def __init__(self, msg, item=None):
        if item is not None:
            msg = '%s: %r' % (msg, item)
        super(Exception, self).__init__(msg)


class ConnectionDeadError(ClientException):
    """Raised when can not connect to the server"""


class ValidationException(ClientException):
    """Raised when an invalid parameter is passed to a ``Client`` function."""



import logging
import tornado.ioloop
import socket
import binascii
from tornado import gen
from toro import Queue, Full, Empty

from .host import Host
from . import constants as const
from .exceptions import ValidationException, ConnectionDeadError


class ConnectionPool(object):

    def __init__(self, servers, maxsize=15, minsize=1, loop=None, debug=0):
        loop = loop if loop is not None else tornado.ioloop.IOLoop.instance()
        if debug:
            logging.basicConfig(
                level=logging.DEBUG,
                format="'%(levelname)s %(asctime)s"
                " %(module)s:%(lineno)d %(process)d %(thread)d %(message)s'"
            )
        self._loop = loop
        self._servers = servers
        self._minsize = minsize
        self._debug = debug
        self._in_use = set()
        self._pool = Queue(maxsize, io_loop=self._loop)

    @gen.coroutine
    def clear(self):
        """Clear pool connections."""
        while not self._pool.empty():
            conn = yield self._pool.get()
            conn.close_socket()

    def size(self):
        return len(self._in_use) + self._pool.qsize()

    @gen.coroutine
    def acquire(self):
        """Acquire connection from the pool, or spawn new one
        if pool maxsize permits.

        :return: ``Connetion`` (reader, writer)
        """
        while self.size() < self._minsize:
            _conn = yield self._create_new_conn()
            yield self._pool.put(_conn)

        conn = None
        while not conn:
            if not self._pool.empty():
                conn = yield self._pool.get()

            if conn is None:
                conn = yield self._create_new_conn()

        self._in_use.add(conn)
        raise gen.Return(conn)

    @gen.coroutine
    def _create_new_conn(self):
        conn = yield Connection.get_conn(self._servers, self._debug)
        raise gen.Return(conn)

    def release(self, conn):
        self._in_use.remove(conn)
        try:
            self._pool.put_nowait(conn)
        except (Empty, Full):
            conn.close_socket()


class Connection(object):

    def __init__(self, servers, debug=0):
        assert isinstance(servers, list)
        self.hosts = [Host(s, self, debug) for s in servers]

    @classmethod
    @gen.coroutine
    def get_conn(cls, servers, debug=0):
        return cls(servers, debug=debug)

    @gen.coroutine
    def send_cmd_all(self, cmd, *arg, **kw):
        res = []
        for host in self.hosts:
            try:
                server_resp = yield host.send_cmd(cmd, *arg, **kw)
                res.append(server_resp)
            except ConnectionDeadError as msg:
                host.mark_dead(msg)
            except socket.error as msg:
                if isinstance(msg, tuple):
                    msg = msg[1]
                host.mark_dead(msg)
        if not len(res):
            raise ConnectionDeadError(
                'no alive connetions {}'.format(
                    ', '.join(
                        map(lambda h: h.disconect_reason, self.hosts)
                    )
                )
            )
        raise gen.Return(res)

    @gen.coroutine
    def send_cmd(self, cmd, *arg, **kw):
        res = yield self.hosts[self._cmemcache_hash(cmd) % len(self.hosts)] \
            .send_cmd(cmd, *arg, **kw)
        raise gen.Return(res)

    def _cmemcache_hash(self, key):
        if isinstance(key, str):
            try:
                key = key.encode('utf-8')
            except UnicodeDecodeError as e:
                raise ValidationException('Hash exception', e)
        try:
            res = (
                (((
                    binascii.crc32(key) & 0xffffffff
                ) >> 16) & 0x7fff) or 1
            )
        except Exception as e:
            raise ValidationException('Hash exception', e)
        return res

    def _get_server(self, key):
        if isinstance(key, tuple):
            serverhash, key = key
        else:
            serverhash = self._cmemcache_hash(key)

        if not self.hosts:
            return None, None

        for i in range(const.SERVER_RETRIES):
            server = self.hosts[serverhash % len(self.hosts)]
            return server, key
        return None, None

    def get_stream(self, cmd, *arg, **kw):
        hosts = self.hosts[self._cmemcache_hash(cmd) % len(self.hosts)] \
            ._ensure_connection()
        return hosts.stream

    def close_socket(self):
        for host in self.hosts:
            host.close_socket()



from asyncmc.pool import Connection
from ._testutil import BaseTest, run_until_complete


class HostTests(BaseTest):
    def setUp(self):
        super(HostTests, self).setUp()

    def tearDown(self):
        super(HostTests, self).tearDown()

    @run_until_complete
    def test_hash_funtion(self):
        conn = Connection(servers=[
            'localhost:11211',
            'some_another_host:11211'
        ])
        s1, _ = conn._get_server('1')
        s2, _ = conn._get_server(b'1')
        self.assertEqual(s1, s2)

        s3, _ = conn._get_server('12')
        self.assertNotEqual(s1, s3)
        conn.close_socket()



STORED = b'STORED'
NOT_STORED = b'NOT_STORED'
TOUCHED = b'TOUCHED'
NOT_FOUND = b'NOT_FOUND'
DELETED = b'DELETED'
VERSION = b'VERSION'
OK = b'OK'
FLAG_PICKLE = 1 << 0
FLAG_INTEGER = 1 << 1
FLAG_JSON = 1 << 2
FLAG_BOOLEAN = 1 << 3
FLAG_STRING = 1 << 4
SERVER_RETRIES = 5
SOCKET_TIMEOUT = 3
DEAD_RETRY = 3



from functools import wraps
import tornado.ioloop
from tornado import gen
import unittest


def run_until_complete(fun):

    @wraps(fun)
    def wrapper(test, *args, **kw):
        @gen.coroutine
        def out():
            return fun(test, *args, **kw)
        test.loop._stopped = False
        test.loop.run_sync(out)
    return wrapper


class BaseTest(unittest.TestCase):

    def setUp(self):
        self.loop = tornado.ioloop.IOLoop.instance()

    def tearDown(self):
        self.loop.stop()
        del self.loop



# -*- coding:utf-8 -*-
from time import sleep
from tornado import gen

from asyncmc.client import Client
from asyncmc.exceptions import ClientException, ValidationException
from ._testutil import BaseTest, run_until_complete


class FooStruct(object):

    def __init__(self):
        self.bar = "baz"

    def __str__(self):
        return "A FooStruct"

    def __eq__(self, other):
        if isinstance(other, FooStruct):
            return self.bar == other.bar
        return 0


class ConnectionCommandsTest(BaseTest):
    def setUp(self):
        super(ConnectionCommandsTest, self).setUp()
        self.mcache = Client(servers=[
            'localhost:11211'
        ], debug=1)
        self.mcache.flush_all(noreply=True)

    def tearDown(self):
        super(ConnectionCommandsTest, self).tearDown()
        self.mcache.close()

    @run_until_complete
    def test_utils(self):
        sleep(1)

    @gen.coroutine
    def check_setget(self, key, val, noreply=False):
        yield self.mcache.set(key, val, noreply=noreply)
        newval = yield self.mcache.get(key)
        self.assertEqual(newval, val)

    @run_until_complete
    def test_version(self):
        version = yield self.mcache.version()
        stats = yield self.mcache.stats()
        self.assertEqual(version, stats[b'version'])

    @run_until_complete
    def test_replace(self):
        key, value = b'key:replace', b'1'
        yield self.mcache.set(key, value)

        key_str, value_str = 'key1:replace', '1'
        yield self.mcache.set(key_str, value_str)

        key_noreply, value_noreply = 'key2:replace', '1'
        yield self.mcache.set(key_noreply, value_noreply)

        test_value = yield self.mcache.replace(key_noreply, '2', noreply=True)
        self.assertEqual(test_value, True)

        test_value = yield self.mcache.replace(key, b'2')
        self.assertEqual(test_value, True)

        test_value = yield self.mcache.get(key)
        self.assertEqual(test_value, b'2')

        test_value = yield self.mcache.replace(key_str, '2')
        self.assertEqual(test_value, True)

        test_value = yield self.mcache.get(key_str)
        self.assertEqual(test_value, '2')

        test_value = yield self.mcache.get(key_noreply)
        self.assertEqual(test_value, '2')

        test_value = yield self.mcache.replace(b'not:' + key, b'3')
        self.assertEqual(test_value, False)

        test_value = yield self.mcache.get(b'not:' + key)
        self.assertEqual(test_value, None)

    @run_until_complete
    def test_append(self):
        key, value = b'key:append', b'1'
        yield self.mcache.set(key, value)
        values = [{
            'val': 1,
            'add': 2,
            'res': 3,
            'res1': 5
        }, {
            'val': '1',
            'add': '2',
            'res': '12',
            'res1': '122'
        }, {
            'val': [1, 2],
            'add': [3, 4],
            'res': [1, 2, 3, 4],
            'res1': [1, 2, 3, 4, 3, 4]
        }]
        for index, val in enumerate(values):
            val['key'] = 'key:append{}'.format(index)
            yield self.mcache.set(val['key'], val['val'])

        app_results = yield [
            self.mcache.append(val['key'], val['add']) for val in values
        ]

        for res in app_results:
            self.assertEqual(res, True)

        app_results = yield [
            self.mcache.get(val['key']) for val in values
        ]

        no_rep_app_results = yield [
            self.mcache.append(
                val['key'],
                val['add'],
                noreply=True
            ) for val in values
        ]

        for res in no_rep_app_results:
            self.assertEqual(res, True)

        for index, res in enumerate(app_results):
            self.assertEqual(res, values[index]['res'])

        test_value = yield self.mcache.append(key, b'2')
        self.assertEqual(test_value, True)

        # make sure value exists
        test_value = yield self.mcache.get(key)
        self.assertEqual(test_value, b'12')

        test_value = yield self.mcache.append(b'not:' + key, b'3')
        self.assertEqual(test_value, False)
        # make sure value exists
        test_value = yield self.mcache.get(b'not:' + key)
        self.assertEqual(test_value, None)

        app_results = yield [
            self.mcache.get(val['key']) for val in values
        ]

        for index, res in enumerate(app_results):
            self.assertEqual(res, values[index]['res1'])

    @run_until_complete
    def test_prepend(self):
        key, value = b'key:prepend', b'1'
        yield self.mcache.set(key, value)

        values = [{
            'val': 1,
            'add': 2,
            'res': 3,
            'res1': 5
        }, {
            'val': '1',
            'add': '2',
            'res': '21',
            'res1': '221'
        }, {
            'val': [1, 2],
            'add': [3, 4],
            'res': [3, 4, 1, 2],
            'res1': [3, 4, 3, 4, 1, 2]
        }]
        for index, val in enumerate(values):
            val['key'] = 'key:prepend{}'.format(index)
            yield self.mcache.set(val['key'], val['val'])

        app_results = yield [
            self.mcache.prepend(val['key'], val['add']) for val in values
        ]

        for res in app_results:
            self.assertEqual(res, True)

        app_results = yield [
            self.mcache.get(val['key']) for val in values
        ]

        for index, res in enumerate(app_results):
            self.assertEqual(res, values[index]['res'])

        no_rep_app_results = yield [
            self.mcache.prepend(
                val['key'],
                val['add'],
                noreply=True
            ) for val in values
        ]

        for res in no_rep_app_results:
            self.assertEqual(res, True)

        test_value = yield self.mcache.prepend(key, b'2')
        self.assertEqual(test_value, True)

        # make sure value exists
        test_value = yield self.mcache.get(key)
        self.assertEqual(test_value, b'21')

        test_value = yield self.mcache.prepend(b'not:' + key, b'3')
        self.assertEqual(test_value, False)
        # make sure value exists
        test_value = yield self.mcache.get(b'not:' + key)
        self.assertEqual(test_value, None)

        app_results = yield [
            self.mcache.get(val['key']) for val in values
        ]

        for index, res in enumerate(app_results):
            self.assertEqual(res, values[index]['res1'])

    @run_until_complete
    def test_add(self):
        key, value = b'key:add', b'1'
        yield self.mcache.set(key, value)

        key_str, value_str = 'key:add_str', '1'
        yield self.mcache.set(key_str, value_str)

        key_noreply, value_noreply = 'key:add_noreply', '1'
        yield self.mcache.set(key_noreply, value_noreply, noreply=True)

        test_value = yield self.mcache.add(key, b'2')
        self.assertEqual(test_value, False)

        test_value = yield self.mcache.add(key_str, b'2')
        self.assertEqual(test_value, False)

        test_value = yield self.mcache.add(key_noreply, b'2', noreply=True)
        # no sense here
        self.assertEqual(test_value, True)

        test_value = yield self.mcache.add('not:' + key_str, '2')
        self.assertEqual(test_value, True)

        test_value = yield self.mcache.get('not:' + key_str)
        self.assertEqual(test_value, '2')

        test_value = yield self.mcache.add(b'not:' + key, b'2')
        self.assertEqual(test_value, True)

        test_value = yield self.mcache.get(b'not:' + key)
        self.assertEqual(test_value, b'2')

    @run_until_complete
    def test_set_typed(self):
        init_val = {
            b'array': [],
            b'dict': {'a': 'b'},
            b'init': 12313,
            'init_str': 12313,
            b'boolean': False,
            b'custom_type': set([1, 4, 4]),
            b'custom_type1': FooStruct()
        }

        results = yield [
            self.mcache.set(key, value)
            for key, value in init_val.items()
        ]

        for res in results:
            self.assertEqual(res, True)

        values = yield dict([
            (key, self.mcache.get(key)) for key, value in init_val.items()
        ])
        for key, val in init_val.items():
            self.assertEqual(val, values[key])

    @run_until_complete
    def test_flush_all(self):
        key, value = b'key:flush_all', b'flush_all_value'
        yield self.mcache.set(key, value)
        # make sure value exists
        test_value = yield self.mcache.get(key)
        self.assertEqual(test_value, value)
        # flush data
        yield self.mcache.flush_all()
        # make sure value does not exists
        test_value = yield self.mcache.get(key)

    @run_until_complete
    def test_set_expire(self):
        key, value = b'key:set', b'1'
        yield self.mcache.set(key, value, exptime=1)
        test_value = yield self.mcache.get(key)
        self.assertEqual(test_value, value)
        sleep(2)
        test_value = yield self.mcache.get(key)
        self.assertEqual(test_value, None)
        self.assertEqual(test_value, None)

    @run_until_complete
    def test_exeptions(self):
        f = False
        try:
            yield self.mcache.set(u'ва', 'sdfadsf')
        except ValidationException:
            f = True
        self.assertEqual(f, True)
        f = False
        try:
            yield self.mcache.multi_get('ecxeption', 'ecxeption')
        except ClientException:
            f = True
        self.assertEqual(f, True)

    @run_until_complete
    def test_set_get(self):
        key, value = b'key:set', b'1'
        yield self.mcache.set(key, value)
        test_value = yield self.mcache.get(key)
        self.assertEqual(test_value, value)
        test_value = yield self.mcache.get(b'not:' + key)
        self.assertEqual(test_value, None)
        yield [
            self.check_setget(
                "a_string",
                "some random string"
            ),
            self.check_setget(
                "a_string_2",
                "some random string",
                noreply=True
            ),
            self.check_setget(
                "an_integer",
                42
            ),
            self.check_setget(
                "an_integer_2",
                42,
                noreply=True
            )
        ]

    @run_until_complete
    def test_delete(self):
        key_noreply, value_noreply = 'key:no_repleay', 'value_norep'
        yield self.check_setget(key_noreply, value_noreply)

        key, value = b'key:delete', b'value'
        yield self.check_setget(key, value)

        key_str, value_str = 'key:delete1', 'value_str'
        yield self.check_setget(key_str, value_str)

        is_deleted = yield self.mcache.delete(key_noreply, noreply=True)
        self.assertTrue(is_deleted)

        is_deleted = yield self.mcache.delete(key)
        self.assertTrue(is_deleted)

        is_deleted = yield self.mcache.delete(key_str)
        self.assertTrue(is_deleted)

        # make sure value does not exists
        test_value = yield self.mcache.get(key)
        self.assertEqual(test_value, None)

        test_value = yield self.mcache.get(key_str)
        self.assertEqual(test_value, None)

        test_value = yield self.mcache.get(key_noreply)
        self.assertEqual(test_value, None)

    @run_until_complete
    def test_str_get(self):
        key1, value1 = 'key:multi_get:1', b'1'
        key2, value2 = 'key:multi_get:2', b'2'
        yield self.mcache.set(key1, value1)
        yield self.mcache.set(key2, value2)
        test_value = yield self.mcache.multi_get(key1, key2)
        self.assertEqual(test_value, [value1, value2])

        test_value = yield self.mcache.get(key1)
        test_value = yield self.mcache.multi_get('not' + key1, key2)
        self.assertEqual(test_value, [None, value2])
        test_value = yield self.mcache.multi_get()
        self.assertEqual(test_value, [])

    @run_until_complete
    def test_multi_get(self):
        key1, value1 = b'key:multi_get:1', b'1'
        key2, value2 = b'key:multi_get:2', b'2'
        yield self.mcache.set(key1, value1)
        yield self.mcache.set(key2, value2)
        test_value = yield self.mcache.multi_get(key1, key2)
        self.assertEqual(test_value, [value1, value2])

        test_value = yield self.mcache.multi_get(b'not' + key1, key2)
        self.assertEqual(test_value, [None, value2])
        test_value = yield self.mcache.multi_get()
        self.assertEqual(test_value, [])

    @run_until_complete
    def test_incr(self):
        key = b'key1'
        yield self.mcache.set(key, 0)
        yield self.mcache.incr(key)
        found_value = int((yield self.mcache.get(key)))
        self.assertEqual(found_value, 1)

    @run_until_complete
    def test_decr(self):
        key = b'key1'
        yield self.mcache.set(key, 2)
        yield self.mcache.decr(key)
        found_value = int((yield self.mcache.get(key)))
        self.assertEqual(found_value, 1)



import socket
import time
# import logging
import tornado.ioloop
import tornado.iostream
from tornado import gen
from . import constants
from . import exceptions


class Host(object):

    def __init__(self, host, conn, debug=0):
        self.debug = debug
        self.host = host
        self.port = 11211
        self.flush_on_reconnect = 1
        self.stream = None

        self.flush_on_next_connect = 0
        self.dead_retry = constants.DEAD_RETRY
        self.deaduntil = 0

        if ":" in self.host:
            parts = self.host.split(":")
            self.host = parts[0]
            self.port = int(parts[1])

        self.sock = None

    def _ensure_connection(self):
        if self.sock:
            return self

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if hasattr(s, 'settimeout'):
            s.settimeout(constants.SOCKET_TIMEOUT)
        try:
            s.connect((self.host, self.port))
        except socket.timeout as msg:
            self.mark_dead('connect: {}'.format(msg))
            return None
        except socket.error as msg:
            if isinstance(msg, tuple):
                msg = msg[1]
            self.mark_dead('connect: {}'.format(msg))
            return None
        self.sock = s
        self.stream = tornado.iostream.IOStream(s)
        self.stream.debug = True
        return self

    def _check_dead(self):
        if self.deaduntil and self.deaduntil > time.time():
            return 1
        self.deaduntil = 0
        return 0

    def mark_dead(self, reason):
        self.disconect_reason = str(reason)
        self.deaduntil = time.time() + self.dead_retry
        if self.flush_on_reconnect:
            self.flush_on_next_connect = 1
        self.close_socket()

    def close_socket(self):
        if self.sock:
            self.stream.close()
            self.sock.close()
            self.sock = None

    @gen.coroutine
    def send_cmd(self, cmd, noreply=False, stream=False):
        self._ensure_connection()
        cmd = cmd + "\r\n".encode()
        if stream:
            yield self.stream.write(cmd)
            raise gen.Return(self.stream)
        elif self.stream:
            yield self.stream.write(cmd)
        if not noreply and self.stream:
            response = yield self.stream.read_until(b'\r\n')
            raise gen.Return(response[:-2])
        if hasattr(self, 'disconect_reason'):
            raise exceptions.ConnectionDeadError(
                'socket host "{}" port "{}" disconected because "{}"'.format(
                    self.host,
                    self.port,
                    self.disconect_reason
                )
            )



from time import sleep
from ._testutil import run_until_complete, BaseTest
from asyncmc.pool import ConnectionPool, Connection
from asyncmc.exceptions import ConnectionDeadError


class PoolTest(BaseTest):

    @run_until_complete
    def test_utils(self):
        sleep(1)
        self.assertEqual(0, 0)

    @run_until_complete
    def test_pool_creation(self):
        pool = ConnectionPool(['localhost:11211'], debug=1)
        self.assertEqual(pool.size(), 0)

    @run_until_complete
    def test_pool_acquire_release(self):
        pool = ConnectionPool(['localhost:11211'], debug=1)
        conn = yield pool.acquire()
        self.assertIsInstance(conn, Connection)
        pool.release(conn)

    @run_until_complete
    def test_pool_clear(self):
        pool = ConnectionPool(['localhost:11211'], debug=1)
        conn = yield pool.acquire()
        pool.release(conn)
        self.assertEqual(pool.size(), 1)
        yield pool.clear()
        self.assertEqual(pool._pool.qsize(), 0)

    @run_until_complete
    def test_pool_half_connection(self):
        pool = ConnectionPool([
            'localhost:11211',
        ])
        conn = yield pool.acquire()
        yield conn.send_cmd_all(b'!')
        pool.release(conn)

    @run_until_complete
    def test_pool_conntection_excetion(self):
        pool = ConnectionPool(
            ['some_host:1233123']
        )
        conn = yield pool.acquire()
        with self.assertRaises(ConnectionDeadError):
            yield conn.send_cmd_all(b'!')
            pool.release(conn)

    @run_until_complete
    def test_pool_is_full(self):
        pool = ConnectionPool(
            ['localhost:11211'],
            minsize=1,
            maxsize=2,
            debug=1
        )
        conn = yield pool.acquire()

        # put garbage to the pool make it look like full
        mocked_conns = [Connection([]), Connection([])]
        yield pool._pool.put(mocked_conns[0])
        yield pool._pool.put(mocked_conns[1])

        # try to return connection back
        self.assertEqual(pool.size(), 3)
        pool.release(conn)
        self.assertEqual(pool.size(), 2)



"""memcached client, based on mixpanel's memcache_client library

Usage example::

    import aiomcache
    mc = aiomcache.Client("127.0.0.1", 11211, timeout=1, connect_timeout=5)
    yield from mc.set("some_key", "Some value")
    value = yield from mc.get("some_key")
    yield from mc.delete("another_key")
"""

from .client import Client
from .exceptions import ClientException, ValidationException

__all__ = ('Client', 'ClientException', 'ValidationException')



import tornado.ioloop
import functools
import pickle
import json
import re
import logging
from tornado import gen

from . import constants as const
from .exceptions import ClientException, ValidationException
from .pool import ConnectionPool

"""client module for memcached (memory cache daemon)

Overview
========

See U{the MemCached homepage<http://www.danga.com/memcached>} for more
about memcached.

Usage summary
=============

This should give you a feel for how this module operates::

    @gen.coroutine
    def out():
        mc = asyncmc.Client(servers=['localhost:11211'], loop=i_loop)
        yield mc.set(b"some_key", b"Some value")
        value = yield mc.get(b"some_key")
        print(value)
        values = yield mc.multi_get(b"some_key", b"other_key")
        print(values)
        yield mc.delete(b"another_key")

    i_loop.run_sync(out)

The standard way to use memcache with a database is like this:

    obj = yield mc.get(key)
    if not obj:
        obj = backend_api.get(...)
        yield mc.set(key, obj)

    # we now have obj, and future passes through this code
    # will use the object from the cache.

Detailed Documentation
======================

More detailed documentation is available in the L{Client} class.

"""


def acquire(func):

    @gen.coroutine
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        conn = yield self.pool.acquire()
        try:
            res = yield func(self, conn, *args, **kwargs)
            raise gen.Return(res)
        finally:
            self.pool.release(conn)

    return wrapper


class Client(object):
    """Object representing a memcache server.

    See L{memcache} for an overview.
    In any case key will be a simple hashable type (string, integer, etc.)


    @group Insertion: set, add, replace
    @group Retrieval: get, get_multi
    @group Removal: delete

    """

    def __init__(self, **kwargs):
        self.debug = kwargs.get('debug')
        self.io_loop = kwargs.get('loop', tornado.ioloop.IOLoop.instance())
        self.pool = ConnectionPool(
            kwargs.get('servers', ["localhost:11211"]),
            debug=self.debug,
            loop=self.io_loop,
            minsize=kwargs.get('pool_minsize', 1),
            maxsize=kwargs.get('pool_size', 15)
        )

        """Create a new Client object with the given list of servers.
            @param servers: C{servers} is passed to L{set_servers}.
            @param loop: Event loop which is used for ansync operations.
                It is optional but if is not defined it will be tornado
                singletone event loop instance.
            @param pool_minsize: Minimal number of connetions with memcashed
                server
            @param pool_size: Maximal number of connetions with memcashed
                server
        """

    # key supports ascii sans space and control chars
    # \x21 is !, right after space, and \x7e is -, right before DEL
    # also 1 <= len <= 250 as per the spec
    _valid_key_re = re.compile(b'^[\x21-\x7e]{1,250}$')

    @acquire
    @gen.coroutine
    def stats(self, conn, args=None):
        """Runs a stats command on the server

        @param args: Additional arguments to pass to the memcache
            "stats" command.
        """
        # req  - stats [additional args]\r\n
        # resp - STAT <name> <value>\r\n (one per result)
        #        END\r\n
        if args is None:
            args = b''
        cmd = b''.join((b'stats ', args))
        resp = yield conn.send_cmd(cmd)
        result = {}
        while resp != b'END\r\n':
            terms = resp.split()

            if len(terms) == 2 and terms[0] == b'STAT':
                result[terms[1]] = None
            elif len(terms) == 3 and terms[0] == b'STAT':
                result[terms[1]] = terms[2]
            else:
                raise ClientException('stats failed', resp)

            resp = yield conn.get_stream(cmd).read_until(b'\r\n')

        raise gen.Return(result)

    @acquire
    @gen.coroutine
    def version(self, conn):
        """Current version of the server.

        @return: bytes, memcached version for current the server.
        """
        command = b'version'
        response = yield conn.send_cmd(command)
        if not response.startswith(const.VERSION):
            raise ClientException('Memcached version failed', response)
        version, number = response.split()
        raise gen.Return(number)

    def _key_type(self, key_list=[], key=None):
        out_keys = []

        if not key_list and key:
            if not isinstance(key, bytes):
                key = key.encode('utf-8')
            return key

        for key in key_list:
            if not isinstance(key, bytes):
                key = key.encode('utf-8')
            out_keys.append(key)
        return out_keys

    def close(self):
        self.pool.clear()

    def _value_type(self, value):
        flag = 0
        if isinstance(value, bytes):
            pass
        elif isinstance(value, str):
            flag |= const.FLAG_STRING
            value = value.encode('utf-8')
        elif isinstance(value, bool):
            flag |= const.FLAG_BOOLEAN
            value = str(int(value)).encode('utf-8')
        elif isinstance(value, int):
            flag |= const.FLAG_INTEGER
            value = str(value).encode('utf-8')
        else:
            try:
                value = json.dumps(value).encode('utf-8')
                flag |= const.FLAG_JSON
            except Exception as e:
                logging.info(e)
                value = pickle.dumps(value, 2)
                flag |= const.FLAG_PICKLE

        if not isinstance(value, bytes):
            logging.info(value)
            value = bytes(value)

        return value, flag

    @acquire
    @gen.coroutine
    def multi_get(self, conn, *keys):
        """Retrieves multiple keys from the memcache doing just one query.

        This method is recommended over regular L{get} as it lowers
        the number of total packets flying around your network,
        reducing total latency, since your app doesn't have to wait
        for each round-trip of L{get} before sending the next one.

        @param keys: list keys for the item being fetched.
        @return: list of values for the specified keys.
        @raises: ValidationException, ClientException,
            and socket errors
        """
        result = yield self._multi_get(conn, *self._key_type(key_list=keys))
        raise gen.Return(result)

    @acquire
    @gen.coroutine
    def flush_all(self, conn, noreply=False):
        """Its effect is to invalidate all existing items immediately"""
        command = b'flush_all' + (b' noreply' if noreply else b'')
        response = yield conn.send_cmd_all(command)

        if [const.OK for n in range(len(response))] != response:
            raise ClientException('Memcached flush_all failed', response)

    @acquire
    @gen.coroutine
    def get(self, conn, key, default=None):
        """Gets a single value from the server.

        @param key: bytes or string, is the key for the item being fetched
        @param default: default value if there is no value.
            #DOTO test default value
        @return: custom type, is the data for this specified key.
        """
        result = yield self._multi_get(conn, self._key_type(key=key))
        result = result[0] if result else default
        raise gen.Return(result)

    @acquire
    @gen.coroutine
    def set(self, conn, key, value, exptime=0, noreply=False):
        """Sets a key to a value on the server
        with an optional exptime (0 means don't auto-expire)

        @param key: bytes or string, is the key of the item.
        @param value: custom type, data to store.
        @param exptime: Tells memcached the time which this value should
            expire, either as a delta number of seconds, or an absolute
            unix time-since-the-epoch value. See the memcached protocol
            docs section "Storage Commands" for more info on <exptime>. We
            default to 0 == cache forever.
        @param noreply: optional parameter instructs the server to not
        send the reply.

        @return: bool, True in case of success.
        """
        resp = yield self._storage_command(
            conn, b'set', self._key_type(key=key), value, exptime, noreply)
        raise gen.Return(resp)

    @gen.coroutine
    def _multi_get(self, conn, *keys):
        # req  - get <key> [<key> ...]\r\n
        # resp - VALUE <key> <flags> <bytes> [<cas unique>]\r\n
        #        <data block>\r\n (if exists)
        #        [...]
        #        END\r\n
        if not keys:
            raise gen.Return([])

        [self._validate_key(key) for key in keys]
        if len(set(keys)) != len(keys):
            raise ClientException('duplicate keys passed to multi_get')
        cmd = b'get ' + b' '.join(keys)
        servers_resp = yield conn.send_cmd_all(cmd, stream=True)
        received = {}
        for stream in servers_resp:
            line = yield stream.read_until(b'\n')
            while line != b'END\r\n':
                terms = line.split()

                if len(terms) == 4 and terms[0] == b'VALUE':  # exists
                    key = terms[1]
                    flags = int(terms[2])
                    length = int(terms[3])

                    val = yield stream.read_bytes(length+2)
                    val = val[:-2]

                    if flags == 0:
                        pass
                    elif flags & const.FLAG_STRING:
                        val = val.decode('utf-8')
                    elif flags & const.FLAG_BOOLEAN:
                        val = bool(int(val))
                    elif flags & const.FLAG_INTEGER:
                        val = int(val)
                    elif flags & const.FLAG_JSON:
                        val = json.loads(val.decode('utf-8'))
                    elif flags & const.FLAG_PICKLE:
                        val = pickle.loads(val)
                    else:
                        val = False

                    if val is False and not flags & const.FLAG_BOOLEAN:
                        raise ClientException('Unknown flag from server')
                    if key in received:
                        raise ClientException('duplicate results from servers')

                    received[key] = val
                else:
                    raise ClientException('get{} failed'.format(cmd), line)
                line = yield stream.read_until(b'\n')

        if len(received) > len(keys):
            raise ClientException('received too many responses')
        res = [received.get(k, None) for k in keys]
        raise gen.Return(res)

    @acquire
    @gen.coroutine
    def replace(self, conn, key, value, exptime=0, noreply=False):
        """Store this data, but only if the server *does*
            already hold data for this key.

        @param key: bytes or string, is the key of the item.
        @param value: custom class,  data to store.
        @param exptime: int is expiration time. If it's 0, the
            item never expires.
        @return: bool, True in case of success.
        """
        res = yield self._storage_command(
            conn, b'replace', self._key_type(key=key), value, exptime, noreply)
        raise gen.Return(res)

    @acquire
    @gen.coroutine
    def append(self, conn, key, value, exptime=0, noreply=False):
        """Add data to an existing key after existing data

        Also see L{prepend}.

        @param key: bytes or string, is the key of the item.
        @param value: custom type, data to store.
        @param exptime: int is expiration time. If it's 0, the
            item never expires.
        @return: bool, True in case of success.
        """
        if isinstance(value, bytes) or isinstance(value, str):
            command = b'append'
        else:
            command = b'set'
            old_val = yield self.get(key)
            value = old_val + value

        res = yield self._storage_command(
            conn, command, self._key_type(key=key), value, exptime, noreply)
        raise gen.Return(res)

    @acquire
    @gen.coroutine
    def prepend(self, conn, key, value, exptime=0, noreply=False):
        """Add data to an existing key before existing data

        @param key: bytes or string, is the key of the item.
        @param value: custom type, data to store.
        @param exptime: int is expiration time. If it's 0, the
            item never expires.
        @return: bool, True in case of success.
        """
        if isinstance(value, bytes) or isinstance(value, str):
            command = b'prepend'
        else:
            command = b'set'
            old_val = yield self.get(key)
            value = value + old_val

        res = yield self._storage_command(
            conn, command, self._key_type(key=key), value, exptime, noreply)
        raise gen.Return(res)

    @acquire
    @gen.coroutine
    def add(self, conn, key, value, exptime=0, noreply=False):
        """Store this data, but only if the server *doesn't* already
        hold data for this key.

        @param noreply: optional parameter instructs the server to not send the
            reply.

        @param value: srings or bytes, data to store.
        @param exptime: int is expiration time. If it's 0, the
            item never expires.
        @return: bool, True in case of success.
        """
        if noreply:
            logging.warning('Call add method with noreply tag')

        res = yield self._storage_command(
            conn, b'add', self._key_type(key=key), value, exptime, noreply)
        raise gen.Return(res)

    @acquire
    @gen.coroutine
    def delete(self, conn, key, noreply=False):
        """Deletes a key/value pair from the server.

        @return: Nonzero on success. True if case values was
             deleted or False to indicate
        that the item with this key was not found.
  .
        @param noreply: optional parameter instructs the server to not send the
            reply.

        """
        server, key = conn._get_server(key)

        key = self._key_type(key=key)
        assert self._validate_key(key)

        command = b'delete ' + key + (b' noreply' if noreply else b'')
        response = yield server.send_cmd(command, noreply)

        if not noreply and response not in (const.DELETED, const.NOT_FOUND):
            raise ClientException('Memcached delete failed', response)
        raise gen.Return(response == const.DELETED or noreply)

    @acquire
    @gen.coroutine
    def incr(self, conn, key, value=1, noreply=False):
        """Increments a key on the server by the given amount.

        If the key doesn't exist or doesn't contain an integer, this command
        will fail.

        @return: The new value on success; otherwise raises ClientException

        @param noreply: optional parameter that instructs the server not to
            send the reply

        Returns
        """
        server, key = conn._get_server(key)

        value = str(value).encode('ascii')

        command = b'incr ' + key + b' ' + value + \
            (b' noreply' if noreply else b'')
        response = yield server.send_cmd(command, noreply)

        if response == const.NOT_FOUND:
            raise ClientException('Key {0} not found'.format(key))
        raise tornado.gen.Return(int(response))

    @acquire
    @gen.coroutine
    def decr(self, conn, key, value=1, noreply=False):
        """Decrements a key on the server by the given amount.

        If the key doesn't exist or doesn't contain an integer, this command
        will fail.

        If you attempt to decrement 0, the new result will be 0.

        @return: The new value on success; otherwise raises ClientException

        @param noreply: optional parameter that instructs the server not to
            send the reply

        Returns
        """
        server, key = conn._get_server(key)

        value = str(value).encode('ascii')

        command = b'decr ' + key + b' ' + value + \
            (b' noreply' if noreply else b'')
        response = yield server.send_cmd(command, noreply)

        if response == const.NOT_FOUND:
            raise ClientException('Key {0} not found'.format(key))
        raise tornado.gen.Return(int(response))

    @gen.coroutine
    def _storage_command(self, conn, command, key, value,
                         exptime=0, noreply=False):
        # req  - set <key> <flags> <exptime> <bytes> [noreply]\r\n
        #        <data block>\r\n
        # resp - STORED\r\n (or others)

        # typically, if val is > 1024**2 bytes server returns:
        #   SERVER_ERROR object too large for cache\r\n
        # however custom-compiled memcached can have different limit
        # so, we'll let the server decide what's too much
        server, key = conn._get_server(key)

        assert self._validate_key(key)

        if not isinstance(exptime, int) or isinstance(exptime, bool):
            raise ValidationException('exptime not int', exptime)
        elif exptime < 0:
            raise ValidationException('exptime negative', exptime)

        value, flags = self._value_type(value)

        args_arr = [flags, exptime, len(value)]
        if noreply:
            args_arr.append('noreply')
        args = [str(a).encode('utf-8') for a in args_arr]
        _cmd = b' '.join([command, key] + args) + b'\r\n'
        cmd = _cmd + value
        server, key = conn._get_server(key)

        resp = yield server.send_cmd(cmd, noreply=noreply)

        if not noreply and resp not in (const.STORED, const.NOT_STORED):
            raise ClientException('stats "{}" failed'.format(cmd), resp)
        raise gen.Return(resp == const.STORED or noreply)

    def _validate_key(self, key):
        if not isinstance(key, bytes):  # avoid bugs subtle and otherwise
            raise ValidationException('key must be bytes', key)

        m = self._valid_key_re.match(key)
        if m:
            # in python re, $ matches either end of line or right before
            # \n at end of line. We can't allow latter case, so
            # making sure length matches is simplest way to detect
            if len(m.group(0)) != len(key):
                raise ValidationException('trailing newline', key)
        else:
            raise ValidationException('invalid key', key)

        return key



import os
import sys
from setuptools import setup, find_packages

version = '0.6.5'

def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()

install_requires = [
    'tornado>=4.1',
    'toro>=0.8'
]

setup(name='asyncmc',
      version=version,
      description=('Minimal pure python tornado memcached client'),
      long_description='\n\n'.join((read('README.rst'), read('CHANGES.rst'))),
      classifiers=[
          'License :: OSI Approved :: MIT License',
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Operating System :: POSIX',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Environment :: Web Environment'],
      author='Eroshenko Dmitriy',
      author_email='erdmko@gmail.com',
      url='https://github.com/ErDmKo/asyncmc/',
      license='MIT',
      packages=find_packages(),
      install_requires = install_requires,
      include_package_data = True)
